/* IMPORT */
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const path = require("path");
const vscode = require("vscode");
const config_1 = require("./config");
const consts_1 = require("./consts");
const document_1 = require("./todo/document");
const utils_1 = require("./utils");
/* CALL TODOS METHOD */
const callTodosMethodOptions = {
    textEditor: undefined,
    checkValidity: true,
    filterer: undefined,
    method: undefined,
    args: [],
    errors: {
        invalid: 'Only todos can perform this action',
        filtered: 'This todo cannot perform this action'
    }
};
function callTodosMethod(options) {
    options = _.merge({}, callTodosMethodOptions, options);
    if (!utils_1.default.editor.isSupported(options.textEditor))
        return;
    const doc = new document_1.default(options.textEditor.document), lines = _.uniq(options.textEditor.selections.map(selection => selection.active.line)), todos = _.filter(lines.map(line => doc.getTodoAt(line, options.checkValidity)));
    if (todos.length !== lines.length)
        vscode.window.showErrorMessage(options.errors.invalid);
    if (!todos.length)
        return;
    const todosFiltered = options.filterer ? todos.filter(options.filterer) : todos;
    if (todosFiltered.length !== todos.length)
        vscode.window.showErrorMessage(options.errors.filtered);
    if (!todosFiltered.length)
        return;
    const edits = _.filter(todosFiltered.map(todo => todo[options.method](...options.args)));
    if (!edits.length)
        return;
    return utils_1.default.editor.applyEdits(options.textEditor, edits);
}
/* COMMANDS */
function archive(textEditor) {
    const doc = textEditor.document, text = doc.getText(), indentation = config_1.default.getKey('indentation'), archiveName = config_1.default.getKey('archive.name'), archiveLabelMatch = _.last(utils_1.default.getAllMatches(textEditor.document.getText(), consts_1.default.regexes.archive)), archiveLabel = archiveLabelMatch ? archiveLabelMatch[0] : `${archiveName}:`, archiveStartIndex = archiveLabelMatch ? archiveLabelMatch.index : -1, archiveEndIndex = archiveStartIndex === -1 ? -1 : archiveStartIndex + archiveLabel.length, archivableText = archiveStartIndex === -1 ? text : text.substr(0, archiveStartIndex), archivableRegexes = [consts_1.default.regexes.todoDone, consts_1.default.regexes.todoCancel], archivableMatches = _.flatten(archivableRegexes.map(re => utils_1.default.getAllMatches(archivableText, re)));
    if (!archivableMatches.length)
        return;
    const archivablePositions = archivableMatches.map(match => doc.positionAt(match.index)), archivableLines = archivablePositions.map(pos => doc.lineAt(pos.line)), archivablePostPositions = archivableMatches.map((match, i) => doc.positionAt(Math.min(text.length - 1, match.index + archivableLines[i].text.length + 1))), // In order to completely remove the line, including the new line
    archivedLines = archivableLines.map(line => `${indentation}${_.trimStart(line.text)}`), archivedText = '\n' + archivedLines.join('\n'), insertText = archiveStartIndex === -1 ? `\n\n${archiveLabel}${archivedText}` : archivedText, insertPos = archiveEndIndex === -1 ? doc.positionAt(text.length - 1) : doc.positionAt(archiveEndIndex), editsRemoveLines = archivableLines.map((line, i) => vscode.TextEdit.delete(new vscode.Range(line.range.start, archivablePostPositions[i]))), editsInsertArchived = vscode.TextEdit.insert(insertPos, insertText), edits = editsRemoveLines.concat(editsInsertArchived);
    return utils_1.default.editor.applyEdits(textEditor, edits);
}
exports.archive = archive;
function start(textEditor) {
    return callTodosMethod({
        textEditor,
        filterer: todo => todo.isBox(),
        method: 'start',
        errors: {
            invalid: 'Only todos can be started',
            filtered: 'Only not already cancelled/done todos can be started'
        }
    });
}
exports.start = start;
function toggleBox(textEditor) {
    return callTodosMethod({
        textEditor,
        checkValidity: false,
        method: 'toggleBox'
    });
}
exports.toggleBox = toggleBox;
function toggleCancel(textEditor) {
    return callTodosMethod({
        textEditor,
        checkValidity: false,
        method: 'toggleCancel'
    });
}
exports.toggleCancel = toggleCancel;
function toggleDone(textEditor) {
    return callTodosMethod({
        textEditor,
        checkValidity: false,
        method: 'toggleDone'
    });
}
exports.toggleDone = toggleDone;
function open() {
    return __awaiter(this, void 0, void 0, function* () {
        const config = config_1.default.get(), { activeTextEditor } = vscode.window, editorPath = activeTextEditor && activeTextEditor.document.uri.fsPath, rootPath = utils_1.default.folder.getRootPath(editorPath);
        if (!rootPath)
            return vscode.window.showErrorMessage('You have to open a project before being able to open its todo file');
        const projectPath = ((yield utils_1.default.folder.getWrapperPathOf(rootPath, editorPath || rootPath, config.file)) || rootPath), todo = utils_1.default.todo.get(projectPath);
        if (!_.isUndefined(todo)) {
            return utils_1.default.file.open(todo.path);
        }
        else {
            const config = config_1.default.get(), defaultPath = path.join(projectPath, config.file);
            yield utils_1.default.file.make(defaultPath, config.defaultContent);
            return utils_1.default.file.open(defaultPath);
        }
    });
}
exports.open = open;
function openEmbedded() {
    return __awaiter(this, void 0, void 0, function* () {
        const files = yield utils_1.default.embedded.getFiles();
        if (!files.length)
            return vscode.window.showErrorMessage('No text files found');
        const regex = utils_1.default.embedded.getRegex(), todos = yield utils_1.default.embedded.getFilesTodos(files, regex), content = yield utils_1.default.embedded.renderTodos(todos);
        if (!content)
            return vscode.window.showInformationMessage('No embedded todos found');
        utils_1.default.editor.open(content);
    });
}
exports.openEmbedded = openEmbedded;
//# sourceMappingURL=commands.js.map