/* IMPORT */
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const absolute = require("absolute");
const findUp = require("find-up");
const fs = require("fs");
const isBinaryPath = require("is-binary-path");
const mkdirp = require("mkdirp");
const path = require("path");
const pify = require("pify");
const string_matches_1 = require("string-matches");
const vscode = require("vscode");
const Commands = require("./commands");
const config_1 = require("./config");
const consts_1 = require("./consts");
/* UTILS */
const Utils = {
    initCommands(context) {
        const { commands, keybindings } = vscode.extensions.getExtension('fabiospampinato.vscode-todo-plus').packageJSON.contributes;
        commands.forEach(({ command, title }) => {
            if (!_.includes(['todo.open', 'todo.openEmbedded'], command))
                return;
            const commandName = _.last(command.split('.')), handler = Commands[commandName], disposable = vscode.commands.registerCommand(command, () => handler());
            context.subscriptions.push(disposable);
        });
        keybindings.forEach(({ command }) => {
            if (_.includes(['todo.open', 'todo.openEmbedded'], command))
                return;
            const commandName = _.last(command.split('.')), disposable = vscode.commands.registerTextEditorCommand(command, Commands[commandName]);
            context.subscriptions.push(disposable);
        });
        return Commands;
    },
    initLanguage() {
        vscode.languages.setLanguageConfiguration(consts_1.default.languageId, {
            wordPattern: /(-?\d*\.\d\w*)|([^\-\`\~\!\#\%\^\&\*\(\)\=\+\[\{\]\}\\\|\;\:\'\"\,\.\<\>\/\?\s]+)/g,
            indentationRules: {
                increaseIndentPattern: consts_1.default.regexes.project,
                decreaseIndentPattern: consts_1.default.regexes.impossible
            }
        });
    },
    getAllMatches(str, regex, multi = true) {
        regex = multi ? new RegExp(regex.source, 'gm') : regex;
        let match, matches = [];
        while (match = regex.exec(str)) {
            matches.push(match);
        }
        return matches;
    },
    matches2ranges(matches) {
        return matches.map(Utils.match2range);
    },
    match2range(match) {
        const first = _.first(match), last = _.last(match), start = match.index + first.indexOf(last), end = start + last.length;
        return { start, end };
    },
    parseGlobs(globs) {
        return `{${_.castArray(globs).join(',')}}`;
    },
    editor: {
        isSupported(textEditor) {
            return textEditor && (textEditor.document.languageId === consts_1.default.languageId);
        },
        makeReplaceEdit(lineNr, replacement, fromCh, toCh) {
            const range = new vscode.Range(lineNr, fromCh, lineNr, toCh || fromCh), replace = vscode.TextEdit.replace(range, replacement);
            return replace;
        },
        applyEdits(textEditor, edits) {
            const uri = textEditor.document.uri, edit = new vscode.WorkspaceEdit();
            edit.set(uri, edits);
            return vscode.workspace.applyEdit(edit);
        },
        open(content) {
            vscode.workspace.openTextDocument({ language: consts_1.default.languageId }).then((textDocument) => {
                vscode.window.showTextDocument(textDocument).then((textEditor) => {
                    textEditor.edit(edit => {
                        const pos = new vscode.Position(0, 0);
                        edit.insert(pos, content);
                        textEditor.document.save();
                    });
                });
            });
        },
        getDoc(file) {
            return __awaiter(this, void 0, void 0, function* () {
                try {
                    return yield vscode.workspace.openTextDocument(file);
                }
                catch (e) { }
            });
        }
    },
    file: {
        open(filepath, isTextDocument = true) {
            filepath = path.normalize(filepath);
            const fileuri = vscode.Uri.file(filepath);
            if (isTextDocument) {
                return vscode.workspace.openTextDocument(fileuri)
                    .then(vscode.window.showTextDocument);
            }
            else {
                return vscode.commands.executeCommand('vscode.open', fileuri);
            }
        },
        read(filepath) {
            return __awaiter(this, void 0, void 0, function* () {
                try {
                    return (yield pify(fs.readFile)(filepath, { encoding: 'utf8' })).toString();
                }
                catch (e) {
                    return;
                }
            });
        },
        readSync(filepath) {
            try {
                return (fs.readFileSync(filepath, { encoding: 'utf8' })).toString();
            }
            catch (e) {
                return;
            }
        },
        make(filepath, content) {
            return __awaiter(this, void 0, void 0, function* () {
                yield pify(mkdirp)(path.dirname(filepath));
                return Utils.file.write(filepath, content);
            });
        },
        write(filepath, content) {
            return __awaiter(this, void 0, void 0, function* () {
                return pify(fs.writeFile)(filepath, content, {});
            });
        }
    },
    folder: {
        getRootPath(basePath) {
            const { workspaceFolders } = vscode.workspace;
            if (!workspaceFolders)
                return;
            const firstRootPath = workspaceFolders[0].uri.fsPath;
            if (!basePath || !absolute(basePath))
                return firstRootPath;
            const rootPaths = workspaceFolders.map(folder => folder.uri.fsPath), sortedRootPaths = _.sortBy(rootPaths, [path => path.length]).reverse(); // In order to get the closest root
            return sortedRootPaths.find(rootPath => basePath.startsWith(rootPath));
        },
        getWrapperPathOf(rootPath, cwdPath, findPath) {
            return __awaiter(this, void 0, void 0, function* () {
                const foundPath = yield findUp(findPath, { cwd: cwdPath });
                if (foundPath) {
                    const wrapperPath = path.dirname(foundPath);
                    if (wrapperPath.startsWith(rootPath)) {
                        return wrapperPath;
                    }
                }
            });
        }
    },
    todo: {
        getFiles(folderPath) {
            const config = config_1.default.get(), { extensions } = vscode.extensions.getExtension('fabiospampinato.vscode-todo-plus').packageJSON.contributes.languages[0], files = _.uniq([config.file, ...extensions]);
            return files.map(file => path.join(folderPath, file));
        },
        get(folderPath) {
            const files = Utils.todo.getFiles(folderPath);
            for (let file of files) {
                const content = Utils.file.readSync(file);
                if (_.isUndefined(content))
                    continue;
                return {
                    path: file,
                    content
                };
            }
        }
    },
    embedded: {
        getRegex() {
            const config = config_1.default.get();
            return new RegExp(config.embedded.regex, 'g');
        },
        getFiles() {
            return __awaiter(this, void 0, void 0, function* () {
                const config = config_1.default.get(), { include, exclude, limit } = config.embedded, files = yield vscode.workspace.findFiles(Utils.parseGlobs(include), Utils.parseGlobs(exclude), limit), filesText = files.filter(file => !isBinaryPath(file.fsPath));
                return filesText;
            });
        },
        getFilesTodos(files, regex) {
            return __awaiter(this, void 0, void 0, function* () {
                const todos = {}; // { [TYPE] => { [FILE] => [{ LINE, NR }] } }
                for (let file of files) {
                    const doc = yield Utils.editor.getDoc(file);
                    if (!doc)
                        continue;
                    const filePath = doc.uri.fsPath;
                    for (let lineNr = 0, lineNrs = doc.lineCount; lineNr < lineNrs; lineNr++) {
                        const line = doc.lineAt(lineNr).text, matches = string_matches_1.default(line, regex);
                        for (let match of matches) {
                            const type = match[1];
                            if (!todos[type])
                                todos[type] = {};
                            if (!todos[type][filePath])
                                todos[type][filePath] = [];
                            todos[type][filePath].push({ line, lineNr });
                        }
                    }
                }
                return todos;
            });
        },
        renderTodos(todos) {
            const config = config_1.default.get(), { indentation, embedded: { groupByFile }, symbols: { box } } = config, lines = [];
            /* LINES */
            const types = Object.keys(todos).sort();
            types.forEach(type => {
                const files = todos[type];
                lines.push(`${type}:`);
                const filePaths = Object.keys(files).sort();
                filePaths.forEach(filePath => {
                    const todos = files[filePath], normalizedFilePath = `/${_.trimStart(filePath, '/')}`;
                    if (groupByFile) {
                        lines.push(`${indentation}@file://${normalizedFilePath}`);
                    }
                    todos.forEach(({ line, lineNr }) => {
                        lines.push(`${indentation}${groupByFile ? indentation : ''}${box} ${_.trimStart(line)} @file://${normalizedFilePath}#${lineNr + 1}`);
                    });
                });
            });
            return lines.length ? `${lines.join('\n')}\n` : '';
        }
    }
};
/* EXPORT */
exports.default = Utils;
//# sourceMappingURL=utils.js.map